""" ``padding`` module.

    see http://www.di-mgt.com.au/cryptopad.html
"""

from wheezy.security.crypto.comp import chr, ord


def pad(s, block_size):
    """Pad with zeros except make the last byte equal to the
    number of padding bytes.

    The convention with this method is usually always to
    add a padding string, even if the original plaintext was
    already an exact multiple of `block_size` bytes.

    ``s`` - byte string.
    """
    n = len(s) % block_size
    if n > 0:
        n = block_size - n
    else:
        n = block_size
    return (chr(0) * (n - 1)).join((s, chr(n)))


def unpad(s, block_size):
    """Strip right by the last byte number.

    ``s`` - byte string.
    """
    n = len(s)
    if n == 0:
        return None
    n = n % block_size
    if n > 0:
        return None
    n = ord(s[-1])
    if n > block_size:
        return None
    return s[:-n]



""" ``authorization`` module.
"""

from wheezy.security.errors import SecurityError


def authorized(wrapped=None, roles=None):
    """Demand the user accessing protected resource is
    authenticated and optionally in one of allowed ``roles``.

    Requires wrapped object to provide attribute principal.

    ``roles`` - a list of authorized roles.

    Here is an example::

        from wheezy.security.principal import Principal

        class Context(object):
            principal = None

            @authorized
            def op_a(self):
                return True

            @authorized(roles=('operator',))
            def op_b(self):
                return True
    """

    def decorate(func):
        if roles:

            def check_roles(context, *args, **kwargs):
                principal = context.principal
                if principal:
                    principal_roles = principal.roles
                    for role in roles:
                        if role in principal_roles:
                            break
                    else:
                        raise SecurityError("Not authorized.")
                    return func(context, *args, **kwargs)
                else:
                    raise SecurityError("Not authorized.")

            return check_roles
        else:

            def check_authenticated(context, *args, **kwargs):
                if context.principal:
                    return func(context, *args, **kwargs)
                else:
                    raise SecurityError("Not authorized.")

            return check_authenticated

    if wrapped is None:
        return decorate
    else:
        return decorate(wrapped)



""" ``principal`` module.
"""


class Principal(object):
    """Container of user specific security information"""

    def __init__(self, id="", roles=(), alias="", extra=""):
        self.id = id
        self.roles = roles
        self.alias = alias
        self.extra = extra

    def dump(self):
        """Dump principal object."""
        return "\x1f".join(
            [self.id, ";".join(self.roles), self.alias, self.extra]
        )

    @classmethod
    def load(cls, s):
        """Load principal object from string."""
        id, roles, alias, extra = s.split("\x1f", 3)
        return cls(id, tuple(roles.split(";")), alias, extra)



""" ``errors`` module.
"""


class SecurityError(RuntimeError):
    """Raised when a security error occurs. It is subclass of
    ``RuntimeError``.
    """

    def __init__(self, message):
        self.message = message



"""
"""

# flake8: noqa

from wheezy.security.authorization import authorized
from wheezy.security.errors import SecurityError
from wheezy.security.principal import Principal

__all__ = ("authorized", "SecurityError", "Principal")
__version__ = "2.0.0"



""" ``comp`` module.
"""

import sys
from os import urandom
from warnings import warn

PY3 = sys.version_info[0] >= 3

if PY3:  # pragma: nocover
    bytes_type = bytes
    str_type = str

    def chr(i):
        return bytes((i,))

    def ord(b):
        return b


else:  # pragma: nocover
    bytes_type = str
    str_type = unicode  # noqa: F821
    chr = chr
    ord = ord


if PY3:  # pragma: nocover

    def n(s, encoding="latin1"):
        if isinstance(s, str_type):
            return s
        return s.decode(encoding)

    def btos(b, encoding):
        return b.decode(encoding)

    def u(s):
        return s


else:  # pragma: nocover

    def n(s, encoding="latin1"):  # noqa
        if isinstance(s, bytes_type):
            return s
        return s.encode(encoding)

    def btos(b, encoding):  # noqa
        return b.decode(encoding)

    def u(s):
        return unicode(s, "unicode_escape")  # noqa: F821


def b(s, encoding="latin1"):  # pragma: nocover
    if isinstance(s, bytes_type):
        return s
    return s.encode(encoding)


# Hash functions
try:  # pragma: nocover
    # Python 2.5+
    from hashlib import md5, sha1, sha224, sha256, sha384, sha512

    def digest_size(d):
        return d().digest_size

    try:
        from hashlib import new as openssl_hash

        def ripemd160():
            return openssl_hash("ripemd160")

        def whirlpool():
            return openssl_hash("whirlpool")

    except ValueError:
        ripemd160 = None
        whirlpool = None
except ImportError:  # pragma: nocover
    import md5  # noqa
    import sha as sha1  # noqa

    sha224 = sha256 = sha384 = sha512 = ripemd160 = whirlpool = None  # noqa

    def digest_size(d):
        return d.digest_size


# Encryption interface


def block_size(c):  # pragma: nocover
    return c.block_size


def encrypt(c, v):  # pragma: nocover
    return c.encrypt(v)


def decrypt(c, v):  # pragma: nocover
    return c.decrypt(v)


# Supported cyphers
aes = None
aes128 = None
aes192 = None
aes256 = None
aes128iv = None
aes192iv = None
aes256iv = None

# Python Cryptography Toolkit (pycrypto, pycryptodome, pycryptodomex)
try:  # noqa pragma: nocover
    try:  # pragma: nocover
        from Cryptodome.Cipher import AES
    except ImportError:  # pragma: nocover
        from Crypto.Cipher import AES

    class AESIVCipher(object):  # pragma: nocover
        """AES cipher that uses random IV for each encrypt operation
        and prepend it to cipher text; decrypt splits input value into
        IV and cipher text.
        """

        block_size = 16

        def __init__(self, key):
            self.key = key

        def encrypt(self, v):
            iv = urandom(16)
            c = AES.new(self.key, AES.MODE_CBC, iv)
            return iv + c.encrypt(v)

        def decrypt(self, v):
            c = AES.new(self.key, AES.MODE_CBC, v[:16])
            return c.decrypt(v[16:])

    # suppored cyphers
    def aes(key, key_size=32):  # pragma: nocover
        assert len(key) >= key_size
        if len(key) < key_size + 16:  # pragma: nocover
            warn("AES%d: key and iv overlap." % (key_size * 8))
        key = key[-key_size:]
        iv = key[:16]
        return lambda: AES.new(key, AES.MODE_CBC, iv)


except ImportError:  # pragma: nocover
    try:  # pragma: nocover
        from cryptography.hazmat.backends import default_backend
        from cryptography.hazmat.primitives.ciphers import (
            Cipher,
            algorithms,
            modes,
        )

        backend = default_backend()

        class AESCipher(object):  # pragma: nocover
            def __init__(self, key, iv):
                alg = algorithms.AES(key)
                self.c = Cipher(alg, modes.CBC(iv), backend=backend)
                self.block_size = alg.block_size

            def encrypt(self, v):
                e = self.c.encryptor()
                return e.update(v) + e.finalize()

            def decrypt(self, v):
                d = self.c.decryptor()
                return d.update(v) + d.finalize()

        class AESIVCipher(object):  # pragma: nocover
            """AES cipher that uses random IV for each encrypt operation
            and prepend it to cipher text; decrypt splits input value into
            IV and cipher text.
            """

            block_size = 16

            def __init__(self, key):
                self.key = key

            def encrypt(self, v):
                iv = urandom(16)
                print("len", len(iv))
                c = AESCipher(self.key, iv)
                return iv + c.encrypt(v)

            def decrypt(self, v):
                # print(v)
                c = AESCipher(self.key, v[:16])
                return c.decrypt(v[16:])

        def aes(key, key_size=32):  # pragma: nocover
            assert len(key) >= key_size
            if len(key) < key_size + 16:  # pragma: nocover
                warn("AES%d: key and iv overlap." % (key_size * 8))
            key = key[-key_size:]
            iv = key[:16]
            return lambda: AESCipher(key, iv)

    except ImportError:  # pragma: nocover
        # TODO: add fallback to other encryption providers
        pass


if aes:

    def aes128(key):
        return aes(key, 16)

    def aes192(key):
        return aes(key, 24)

    def aes256(key):
        return aes(key, 32)

    def aesiv(key, key_size=32):
        assert len(key) >= key_size
        c = AESIVCipher(key[:key_size])
        return lambda: c

    def aes128iv(key):
        return aesiv(key, 16)

    def aes192iv(key):
        return aesiv(key, 24)

    def aes256iv(key):
        return aesiv(key, 32)



""" ``crypto`` module.
"""

from base64 import b64decode, b64encode
from binascii import Error as BinError
from hmac import new as hmac_new
from os import urandom
from struct import pack, unpack
from time import time
from warnings import warn

from wheezy.security.crypto.comp import (
    aes128,
    b,
    block_size,
    btos,
    decrypt,
    digest_size,
    encrypt,
    sha1,
)
from wheezy.security.crypto.padding import pad, unpad

BASE64_ALTCHARS = b("-~")
EMPTY = b("")
EPOCH = 1317212745


def ensure_strong_key(key, digestmod):
    """Translates a given key to a computed strong key of length
    3 * digestmode.digest_size suitable for encryption, e.g.
    with digestmod set to ``sha1`` returns 480 bit (60 bytes) key.
    """
    hmac = hmac_new(key, key, digestmod)
    k1 = hmac.digest()
    hmac.update(k1)
    k2 = hmac.digest()
    hmac.update(k2)
    return k1 + k2 + hmac.digest()


def timestamp():
    return int(time()) - EPOCH


class Ticket(object):
    """Protects sensitive information (e.g. user id).

    Default policy applies verification and encryption.
    Verification is provided by ``hmac`` initialized with ``sha1``
    digestmod. Encryption is provided if available, by default
    it attempts to use AES cypher.
    """

    __slots__ = ("cypher", "max_age", "hmac", "digest_size", "block_size")

    def __init__(
        self, max_age=900, salt="", digestmod=None, cypher=aes128, options=None
    ):
        self.max_age = max_age
        if not digestmod:
            warn(
                "Ticket: digestmod is not specified, fallback to sha1",
                stacklevel=2,
            )
            digestmod = sha1
        options = options or {}
        key = b(salt + options.get("CRYPTO_VALIDATION_KEY", ""))
        key = ensure_strong_key(key, digestmod)
        self.hmac = hmac_new(key, digestmod=digestmod)
        self.digest_size = digest_size(digestmod)
        if cypher:
            key = b(salt + options.get("CRYPTO_ENCRYPTION_KEY", ""))
            key = ensure_strong_key(key, digestmod)
            self.cypher = cypher(key)
            self.block_size = block_size(self.cypher())
        else:
            self.cypher = None
            warn("Ticket: cypher not available", stacklevel=2)

    def encode(self, value, encoding="UTF-8"):
        """Encode ``value`` according to ticket policy."""
        value = b(value, encoding)
        expires = pack("<i", timestamp() + self.max_age)
        noise = urandom(12)
        value = EMPTY.join((noise[:4], expires, noise[4:8], value, noise[8:]))
        if self.cypher:
            value = encrypt(self.cypher(), pad(value, self.block_size))
        return btos(
            b64encode(self.sign(value) + value, BASE64_ALTCHARS), "latin1"
        )

    def decode(self, value, encoding="UTF-8"):
        """Decode ``value`` according to ticket policy."""
        if len(value) < 48:
            return (None, None)
        try:
            value = b64decode(b(value), BASE64_ALTCHARS)
        except (TypeError, BinError):
            return (None, None)
        signature = value[: self.digest_size]
        value = value[self.digest_size :]
        if signature != self.sign(value):
            return (None, None)
        if self.cypher:
            if len(value) % self.block_size != 0:
                return (None, None)
            value = unpad(decrypt(self.cypher(), value), self.block_size)
            if value is None:
                return (None, None)
        if len(value) < 16:  # pragma: nocover
            return (None, None)
        expires, value = value[4:8], value[12:-4]
        time_left = unpack("<i", expires)[0] - timestamp()
        if time_left < 0 or time_left > self.max_age:
            return (None, None)
        try:
            return (btos(value, encoding), time_left)
        except UnicodeDecodeError:
            return (None, None)

    def sign(self, value):
        """Compute hmac digest."""
        h = self.hmac.copy()
        h.update(value)
        return h.digest()
